@@ -19,10 +19,10 @@ A simple filesystem-backed store
 """
 
 import errno
-import hashlib
 import logging
 import os
 import stat
+import urllib
 
 import jsonschema
 from oslo_config import cfg
@@ -30,7 +30,6 @@ from oslo_serialization import jsonutils
 from oslo_utils import encodeutils
 from oslo_utils import excutils
 from oslo_utils import units
-from six.moves import urllib
 
 import glance_store
 from glance_store import capabilities
@@ -71,6 +70,14 @@ Related options:
 
 """),
     cfg.MultiStrOpt('filesystem_store_datadirs',
+                    deprecated_for_removal=True,
+                    deprecated_since='Flamingo',
+                    deprecated_reason="""
+Users willing to use multiple data directories should configure multiple
+filesystem stores instead of using filesystem_store_datadirs.
+
+This option is scheduled for removal in the H development cycle.
+""",
                     help="""
 List of directories and their priorities to which the filesystem
 backend store writes images.
@@ -110,11 +117,13 @@ Related options:
                help="""
 Filesystem store metadata file.
 
-The path to a file which contains the metadata to be returned with
-any location associated with the filesystem store. The file must
-contain a valid JSON object. The object should contain the keys
-``id`` and ``mountpoint``. The value for both keys should be a
-string.
+The path to a file which contains the metadata to be returned with any location
+associated with the filesystem store. Once this option is set, it is used for
+new images created afterward only - previously existing images are not
+affected.
+
+The file must contain a valid JSON object. The object should contain the keys
+``id`` and ``mountpoint``. The value for both keys should be a string.
 
 Possible values:
     * A valid path to the store metadata file
@@ -165,7 +174,29 @@ Possible Values:
 Related options:
     * None
 
-""")]
+"""),
+    cfg.BoolOpt('filesystem_thin_provisioning',
+                default=False,
+                help="""
+Enable or not thin provisioning in this backend.
+
+This configuration option enable the feature of not really write null byte
+sequences on the filesystem, the holes who can appear will automatically
+be interpreted by the filesystem as null bytes, and do not really consume
+your storage.
+Enabling this feature will also speed up image upload and save network traffic
+in addition to save space in the backend, as null bytes sequences are not
+sent over the network.
+
+Possible Values:
+    * True
+    * False
+
+Related options:
+    * None
+
+"""),
+]
 
 MULTI_FILESYSTEM_METADATA_SCHEMA = {
     "type": "array",
@@ -262,6 +293,14 @@ class Store(glance_store.driver.Store):
     OPTIONS = _FILESYSTEM_CONFIGS
     FILESYSTEM_STORE_METADATA = None
 
+    def __init__(self, *args, **kargs):
+        super(Store, self).__init__(*args, **kargs)
+        if self.backend_group:
+            self.store_conf = glance_store.driver.BackendGroupConfiguration(
+                self.OPTIONS, self.backend_group, conf=self.conf)
+        else:
+            self.store_conf = self.conf.glance_store
+
     def get_schemes(self):
         return ('file', 'filesystem')
 
@@ -288,11 +327,7 @@ class Store(glance_store.driver.Store):
         :datadir is a directory path in which glance writes image files.
         """
 
-        if self.backend_group:
-            fstore_perm = getattr(
-                self.conf, self.backend_group).filesystem_store_file_perm
-        else:
-            fstore_perm = self.conf.glance_store.filesystem_store_file_perm
+        fstore_perm = self.store_conf.filesystem_store_file_perm
 
         if fstore_perm <= 0:
             return
@@ -398,17 +433,14 @@ class Store(glance_store.driver.Store):
         this method. If the store was not able to successfully configure
         itself, it should raise `exceptions.BadStoreConfiguration`
         """
-        if self.backend_group:
-            store_conf = getattr(self.conf, self.backend_group)
-        else:
-            store_conf = self.conf.glance_store
+        fdir = self.store_conf.filesystem_store_datadir
+        fdirs = self.store_conf.filesystem_store_datadirs
+        fstore_perm = self.store_conf.filesystem_store_file_perm
+        meta_file = self.store_conf.filesystem_store_metadata_file
 
-        fdir = store_conf.filesystem_store_datadir
-        fdirs = store_conf.filesystem_store_datadirs
-        fstore_perm = store_conf.filesystem_store_file_perm
-        meta_file = store_conf.filesystem_store_metadata_file
-
-        self.chunk_size = store_conf.filesystem_store_chunk_size
+        self.thin_provisioning = self.store_conf.\
+            filesystem_thin_provisioning
+        self.chunk_size = self.store_conf.filesystem_store_chunk_size
         self.READ_CHUNKSIZE = self.chunk_size
         self.WRITE_CHUNKSIZE = self.READ_CHUNKSIZE
 
@@ -713,19 +745,31 @@ class Store(glance_store.driver.Store):
 
         if os.path.exists(filepath):
             raise exceptions.Duplicate(image=filepath)
-        os_hash_value = hashlib.new(str(hashing_algo))
-        checksum = hashlib.md5()
+        os_hash_value = utils.get_hasher(hashing_algo, False)
+        checksum = utils.get_hasher('md5', False)
         bytes_written = 0
         try:
             with open(filepath, 'wb') as f:
                 for buf in utils.chunkreadable(image_file,
                                                self.WRITE_CHUNKSIZE):
+                    actual_to_write = bytes_written + len(buf)
+                    if image_size != 0 and actual_to_write > image_size:
+                        raise glance_store.Invalid(
+                            _("Size exceeds: expected "
+                              "%(expected)d "
+                              "bytes, got %(actual)d bytes") %
+                            {'expected': image_size,
+                             'actual': actual_to_write})
                     bytes_written += len(buf)
                     os_hash_value.update(buf)
                     checksum.update(buf)
                     if verifier:
                         verifier.update(buf)
-                    f.write(buf)
+                    if self.thin_provisioning and not any(buf):
+                        f.truncate(bytes_written)
+                        f.seek(0, os.SEEK_END)
+                    else:
+                        f.write(buf)
         except IOError as e:
             if e.errno != errno.EACCES:
                 self._delete_partial(filepath, image_id)
@@ -737,6 +781,15 @@ class Store(glance_store.driver.Store):
             with excutils.save_and_reraise_exception():
                 self._delete_partial(filepath, image_id)
 
+        # Final size check after reading all chunks
+        if image_size != 0 and bytes_written != image_size:
+            # Cleanup and raise exception after write size mismatch
+            self._delete_partial(filepath, image_id)
+            message = (_("Size mismatch: expected %(expected)d bytes, "
+                         "got %(actual)d bytes") %
+                       {'expected': image_size, 'actual': bytes_written})
+            raise glance_store.Invalid(message=message)
+
         hash_hex = os_hash_value.hexdigest()
         checksum_hex = checksum.hexdigest()
         metadata = self._get_metadata(filepath)
@@ -748,11 +801,7 @@ class Store(glance_store.driver.Store):
                    'checksum_hex': checksum_hex,
                    'hash_hex': hash_hex})
 
-        if self.backend_group:
-            fstore_perm = getattr(
-                self.conf, self.backend_group).filesystem_store_file_perm
-        else:
-            fstore_perm = self.conf.glance_store.filesystem_store_file_perm
+        fstore_perm = self.store_conf.filesystem_store_file_perm
 
         if fstore_perm > 0:
             perm = int(str(fstore_perm), 8)
@@ -764,7 +813,7 @@ class Store(glance_store.driver.Store):
 
         # Add store backend information to location metadata
         if self.backend_group:
-            metadata['backend'] = u"%s" % self.backend_group
+            metadata['store'] = self.backend_group
 
         return ('file://%s' % filepath,
                 bytes_written,
